import gzip
from collections import UserDict
from pathlib import Path
from typing import Dict, Union

import numpy as np
import numpy.typing as npt
from Bio.Data.IUPACData import protein_letters_3to1_extended
protein_letters_3to1_extended = {res_3_letters.upper():res_1_letter for res_3_letters, res_1_letter in protein_letters_3to1_extended.items()} # By default biopython 3 letter codes are cammel case (ie: 'Gly), but they are all upper case (ie: 'GLY') in parsed PDB files
from Bio.PDB import PDBParser

from Bio.PDB.Residue import Residue as Biopython_residue_type
from Bio.PDB.Structure import Structure

from pyscomotif.data_containers import Residue
from pyscomotif.utils import (get_PDB_ID_from_file_path,
                                pickle_and_compress_python_object)


def parse_PDB_with_biopython(PDB_file_path: Path) -> Structure:
    """
    """
    parser = PDBParser(QUIET=True)
    parsed_PDB_file: Structure
    try:    
        if len(PDB_file_path.suffixes) == 1: # PDB file in any of the valid formats (PDB, mmCIF, etc) so Biopython should be able to parse it.
            parsed_PDB_file = parser.get_structure(id='', file=PDB_file_path)
        elif len(PDB_file_path.suffixes) == 2 and PDB_file_path.suffixes[-1] == '.gz':
            with gzip.open(PDB_file_path, 'rt') as decompressed_PDB_file_handle:
                parsed_PDB_file = parser.get_structure(id='', file=decompressed_PDB_file_handle) # The biopython PDB parser also accepts open file handles
        else:
            raise ValueError(f"<{PDB_file_path.suffix}> compressed PDB files are currently not supported, only gunziped (.gz) compressed PDB files are.")

    except Exception as exception:
        raise exception
    
    return parsed_PDB_file

def add_C_alpha_attribute(residue_data: Dict[str, Biopython_residue_type]) -> None:
    """
    """
    residues_to_ignore = []
    for residue_ID, residue in residue_data.items():
        atoms_dict = residue.child_dict
        if not 'CA' in atoms_dict:
            residues_to_ignore.append(residue_ID)
            continue

        residue.C_alpha = atoms_dict['CA'].coord.round(2) # Residue objects are mutable so no need to reasign them to their residue_ID key


    for residue_ID in residues_to_ignore:
        del residue_data[residue_ID]

    return

def add_sidechain_CMR_attribute(residue_data: Dict[str, Biopython_residue_type]) -> None:
    """
    """
    residues_to_ignore = []
    for residue_ID, residue in residue_data.items():
        if residue.resname == 'GLY':
            sidechain_atoms_coords = np.array([residue.C_alpha])
        else:
            atoms_list = residue.child_list
            sidechain_atoms_coords = np.array([
                atom.coord
                for atom in atoms_list
                if atom.name not in {'N', 'CA', 'C', 'O'} and atom.element != 'H' # All atoms except main chain and hydrogens
            ])
        
        if len(sidechain_atoms_coords) == 0:
            residues_to_ignore.append(residue_ID)
            continue

        residue.sidechain_CMR = sidechain_atoms_coords.mean(axis=0).round(2)


    for residue_ID in residues_to_ignore:
        del residue_data[residue_ID]

    return

def calculate_glycine_vector(residue: Biopython_residue_type) -> Union[npt.NDArray[np.float64], None]:
    """
    Given we cannot calculate a vector between the C_alpha and the side chain CMR as there is no sidechain
    coordinate to be used for Glycine, we calculate a vector that points towards the direction of the C_alpha
    by taking the mid-point between the N and C coordinates, which is simply the average of the x/y/z
    coordinates of the two points, and use that to calculate the vector mid-point -> C_alpha.
    """
    atoms_dict = residue.child_dict
    if not ('N' in atoms_dict and 'C' in atoms_dict):
        return None
    
    N_coord: npt.NDArray[np.float64] = atoms_dict['N'].coord
    C_coord: npt.NDArray[np.float64] = atoms_dict['C'].coord
    C_alpha: npt.NDArray[np.float64] = residue.C_alpha

    N_C_midpoint: npt.NDArray[np.float64] = np.array([N_coord, C_coord]).mean(axis=0)
    return C_alpha - N_C_midpoint

def add_C_alpha_sidechain_CMR_vector_attribute(residue_data: Dict[str, Biopython_residue_type]) -> None:
    """
    """
    residues_to_ignore = []
    for residue_ID, residue in residue_data.items():
        vector = residue.sidechain_CMR - residue.C_alpha if residue.resname != 'GLY' else calculate_glycine_vector(residue)
        if vector is None:
            residues_to_ignore.append(residue_ID)
            continue

        residue.vector = vector.round(2)


    for residue_ID in residues_to_ignore:
        del residue_data[residue_ID]

    return

def add_PDB_header_description_as_dict_attribute(residue_data: UserDict[str,Residue], parsed_PDB_file: Structure) -> None:
    """
    """
    header_description: str = parsed_PDB_file.header['name']

    residue_data.header_description = header_description #type: ignore
    
    return

def extract_residue_data(PDB_file_path: Path) -> UserDict[str, Residue]:
    """
    ...
    """
    parsed_PDB_file = parse_PDB_with_biopython(PDB_file_path)

    residue_data = {
        f'{chain.id}{biopython_residue_object.id[1]}':biopython_residue_object # Key example: A43.
        for chain in parsed_PDB_file.get_chains()
            for biopython_residue_object in chain
    }

    add_C_alpha_attribute(residue_data)
    add_sidechain_CMR_attribute(residue_data)
    add_C_alpha_sidechain_CMR_vector_attribute(residue_data)

    # The Biopython Residue objects contain a lot of data we are not interested in, so we replace them with our own Residue object 
    # that only contains the data we actually care about. 
    # We also want to have access to the PDB's header description/title, which is why we use a UserDict which allows us to add that 
    # string as an attribute to the UserDict object instead of a key in the dictionary, which would be confusing as it's not a Residue object.
    customized_residue_data = UserDict({
        residue_id:Residue(resname=protein_letters_3to1_extended[residue.resname], C_alpha=residue.C_alpha, sidechain_CMR=residue.sidechain_CMR, vector=residue.vector)
        for residue_id, residue in residue_data.items()
        if residue.resname in protein_letters_3to1_extended
    })
    add_PDB_header_description_as_dict_attribute(customized_residue_data, parsed_PDB_file)
    
    return customized_residue_data

def get_compressed_residue_data_output_file_path(PDB_file_path: Path, index_folder_path: Path, compression: str) -> Path:
    """
    """
    PDB_ID = get_PDB_ID_from_file_path(PDB_file_path)
    residue_data_folder_path = index_folder_path / 'residue_data_folder'
    compressed_residue_data_output_file_path = residue_data_folder_path / f'{PDB_ID}.{compression}' # Ex: /home/user_name/database_folder/pyScoMotif_index/residue_data_folder/1A2Y.bz2

    return compressed_residue_data_output_file_path

def generate_compressed_residue_data_from_PDB_file(PDB_file_path: Path, index_folder_path: Path, compression: str) -> None:
    """
    """
    residue_data = extract_residue_data(PDB_file_path)
    pickle_and_compress_python_object(
        python_object=residue_data,
        output_file_path=get_compressed_residue_data_output_file_path(PDB_file_path, index_folder_path, compression)
    )

    return



from dataclasses import dataclass
from typing import Any

import numpy as np
import numpy.typing as npt


@dataclass
class Residue():
    """Used to store all the data needed about a residue to then calculate the geometric descriptors of pairs of residues."""
    resname: str # Ex: 'A' (Alanine)
    C_alpha: npt.NDArray[np.float64]
    sidechain_CMR: npt.NDArray[np.float64]
    vector: npt.NDArray[np.float64]

    def __getitem__(self, key: str) -> Any:
        """To enable bracket notation to get an attribute value (Ex: Residue['resname'])"""
        return getattr(self, key)

@dataclass
class Residue_pair_data():
    """Used to store the data associated with a pair of residues. Typically used as a value in a dictionary where the keys are pairs of residues."""
    C_alpha_distance: float
    sidechain_CMR_distance: float
    vector_angle: float


import bz2
import gzip
import itertools
import pickle
from pathlib import Path
from typing import Any, Iterable, List, Tuple, Union, Dict
from threading import BoundedSemaphore
from concurrent.futures import ProcessPoolExecutor, Future

import numba
import numpy as np
import numpy.typing as npt

from pyscomotif.constants import AMINO_ACID_ALPHABET
from pyscomotif.data_containers import Residue


def get_PDB_ID_from_file_path(PDB_file_path: Path) -> str:
    return PDB_file_path.name.split('.')[0]

def get_sorted_2_tuple(tuple_: Tuple[Any, Any]) -> Tuple[Any, Any]:
    return tuple_ if tuple_[0] <= tuple_[1] else (tuple_[1], tuple_[0])

def sort_and_join_2_strings(str1: str, str2: str) -> str:
    return str1+str2 if str1 <= str2 else str2+str1

def flatten_iterable(iterable: Iterable[Any]) -> Iterable[Any]:
    return itertools.chain.from_iterable(iterable)

def pickle_and_compress_python_object(python_object: Any, output_file_path: Path) -> None:
    """
    ...
    """
    if output_file_path.suffix == '.bz2':
        with bz2.open(output_file_path, 'wb') as file_handle:
            pickle.dump(python_object, file_handle)
    elif output_file_path.suffix == '.gz':
        with gzip.open(output_file_path, 'wb') as file_handle:
            pickle.dump(python_object, file_handle)
    else:
        raise ValueError(f"'{output_file_path.suffix}' is currently not supported, only 'bz2' (default) and 'gz' compression are.")

    return

def read_compressed_and_pickled_file(file_path: Path) -> Any:
    """
    """
    python_object: Any
    if file_path.suffix == '.bz2':
        with bz2.open(file_path, 'rb') as file_handle:
            python_object = pickle.load(file_handle)
    elif file_path.suffix == '.gz':
        with gzip.open(file_path, 'rb') as file_handle:
            python_object = pickle.load(file_handle)
    else:
        raise ValueError(f"'{file_path.suffix}' is currently not supported, only 'bz2' (default) and 'gz' compression are.")

    return python_object

def detect_the_compression_algorithm_used_in_the_index(index_folder_path: Path) -> str:
    """
    Detection of the compression algorithm is simply done by taking a file from the residue data folder and checking its extension.
    """
    residue_data_folder = index_folder_path / 'residue_data_folder'
    random_file = next(residue_data_folder.glob('*'))

    compression = random_file.suffix[1:] # We don't want to include the leading dot (.) of the suffix. Ex: '.txt' -> 'txt'
    return compression

def generate_all_pairs_of_residues_combinations() -> List[str]:
    """
    Returns a list of strings that correspond to all the possible pairs of amino acids, in alphabetical order (e.g: AA, AC, AD, ..., YY)
    """
    return [AA1+AA2 for AA1, AA2 in itertools.combinations_with_replacement(sorted(AMINO_ACID_ALPHABET), 2)] # with_replacement needed so that we also get AA, CC, DD, etcetc


@numba.njit() # type: ignore
def get_bin_number(metric_value: float, bin_size: float) -> int:
    return int(metric_value // bin_size)

@numba.njit() # type: ignore
def pairwise_euclidean_distance(v1: npt.NDArray[np.float64], v2: npt.NDArray[np.float64]) -> Union[np.float64, float]:
    return np.sqrt(np.sum((v1 - v2)**2)) # type: ignore

@numba.njit() # type: ignore
def angle_between_two_vectors(v1: npt.NDArray[np.float64], v2: npt.NDArray[np.float64]) -> Union[np.float64, float]:
    """
    Returns the angle between the two vectors in degrees (°). 
    Largely taken from https://stackoverflow.com/questions/2827393/angles-between-two-n-dimensional-vectors-in-python
    """
    v1_u, v2_u = v1 / np.linalg.norm(v1), v2 / np.linalg.norm(v2)
    
    dot_product = np.dot(v1_u, v2_u)
    # Clip the value to between -1 and 1 (numba does not support np.clip)
    if dot_product > 1:
        dot_product = 1
    elif dot_product < -1:
        dot_product = -1

    return np.rad2deg(np.arccos(dot_product)) # type: ignore

def get_C_alpha_distance(residue_1: Residue, residue_2: Residue) -> float:
    return float(pairwise_euclidean_distance(residue_1.C_alpha, residue_2.C_alpha))

def get_sidechain_CMR_distance(residue_1: Residue, residue_2: Residue) -> float:
    return float(pairwise_euclidean_distance(residue_1.sidechain_CMR, residue_2.sidechain_CMR))

def get_vector_angle(residue_1: Residue, residue_2: Residue) -> float:
    return float(angle_between_two_vectors(residue_1.vector, residue_2.vector))
ProcessPoolExecutor()

class BoundedProcessPoolExecutor(ProcessPoolExecutor):
    """
    This class wraps a concurrent.futures.Executor to limit the size of its task queue. The default concurrent.futures.Executor is unbounded, 
    and therefore if hundrends of thousands or millions of tasks are submited in one go, this can crash the program.
    
    All credits for the code go to https://github.com/mowshon/bounded_pool_executor/tree/master.
    """
    def __init__(self, max_workers: int, max_submited_tasks: int) -> None:
        super().__init__(max_workers=max_workers)
        self.semaphore = BoundedSemaphore(value=max_submited_tasks)

    def submit(self, fn: Any, *args, **kwargs) -> Future[Any]: # type:ignore
        self.semaphore.acquire()
        future = super().submit(fn, *args, **kwargs)
        future.add_done_callback(self.release)

        return future
    
    def release(self, _: Future[Any]) -> None:
        """Called once a task is done, releases one queue slot."""
        self.semaphore.release()